from typing import Callable, Generic, Optional
from fpylib.functors.functor import T, S, Functor
from fpylib.functors.monad import Monad, unit, unitifier
from fpylib.functors.applicative import Applicative


class Maybe(Applicative, Monad, Generic[T]):
    """
    This is a implementation of the Maybe Monad of Haskell. It is a functor, applicative and monad.
    """

    def unit(self, value: T) -> "Maybe[T]":
        """
        Return a Just pr Nothing value based on if the value is None or not.

        :param value: The value to be checked.
        :type value: T
        :return: Just value or Nothing
        """
        if value is not None:
            return Just(value)
        return Nothing()

    def bind(self, func: Callable[[T], S]) -> "Maybe[S]":
        """
        Return a Just pr Nothing value based on if occur an error or not.

        :param func: The function to be applied.
        :type func: Callable[[T], S]
        :return: Just value or Nothing
        """
        try:
            value = func(self.get())
            if value is None:
                return Nothing(ValueError("The value is None"))
            return Just(value)
        except Exception as e:
            return Nothing(e)


class Just(Maybe):
    def __str__(self) -> str:
        return f"Just {self.get()}"

    def __repr__(self) -> str:
        return f"Just {type(self.get())}"


class Nothing(Maybe):
    def __init__(self, *failure: Optional[Exception]) -> None:
        """
        This does nothing.
        """
        object.__setattr__(
            self, "_Nothing__failure", filter(lambda fail: fail is not None, failure)
        )

    def fails(self) -> bool:
        return self.__failure

    def __str__(self) -> str:
        return "Nothing"

    def __repr__(self) -> str:
        return f"{self.__str__()} {list(self.__failure)}"


def maybe_conditioner(func: Callable[..., T]) -> Callable[..., "Maybe[T]"]:
    """
    Conditioner for Maybe.

    :param func: The function to wrap in a Monad.
    :type func: Callable[..., T]
    :return: The wrapped function.
    :rtype: Callable[..., Monad[T]]
    """

    def wrapper(*arg, **kwargs) -> "Maybe[T]":
        try:
            return unit(Maybe, func(*arg, **kwargs))
        except Exception as e:
            return Nothing(e)

    return wrapper



from typing import Any, Callable, Generic, Optional
from fpylib.functors.functor import T, S
from fpylib.lazyness import lazy_eval


class Monad(Generic[T]):
    """
    This is a implementation of the Monad law:
    - Unit or Return .
    - Bind (>>=) or (>>) here.
    """

    def __init__(self, value: Optional[T] = None) -> None:
        """
        Initialize the Monad with a value.

        :param value: The value to initialize the Monad with.
        :return: None
        """
        object.__setattr__(self, "_Monad__value", value)

    def __setattr__(self, __name: str, __value: Any) -> None:
        raise AttributeError("This object is not modifiable")

    def get(self) -> T:
        """
        Get the value of the Monad.

        :return: The value of the Monad.
        """
        return self.__value

    def unit(self, value: T) -> "Monad[T]":
        """
        Return a Monad with the given value.

        :param value: The value to return.
        :return: A Monad with the given value.
        """
        return Monad(value)

    def bind(self, func: Callable[[T], S]) -> "Monad[S]":
        """
        Return a Monad with the result of the given function evaluated with the value of the Monad.

        :param func: The function to evaluate with the value of the Monad.
        :return: A Monad with the result of the given function evaluated with the value of the Monad.
        """
        return Monad(func(self.get()))

    def __rshift__(self, func: Callable[[[T]], S]) -> "Monad[S]":
        """
        Use the >> operator to bind the given function to the value of the Monad.

        For example:
        >>> m = Monad(1) >> (lambda x: x + 1)
        >>> m.get()
        2

        :param func: The function to bind to the value of the Monad.
        :return: A Monad with the result of the given function evaluated with the value of the Monad.
        """
        return self.bind(func)


@lazy_eval
def unit(m: "Monad", value: T) -> "Monad[T]":
    """
    The unit function for the Monad.

    :param m: The Monad.
    :param value: The value to be wrapped in the Monad.
    :return: The Monad with the value.
    """
    return m.unit(m, value)


def unitifier(
    m: "Monad",
    conditioner: Optional[
        Callable[[Callable[..., T]], Callable[..., "Monad[T]"]]
    ] = None,
) -> Callable[[Callable[..., T]], Callable[..., "Monad[T]"]]:
    """
    Return a decorator that wraps the given function in a Monad.
    :param m: The Monad.
    :type m: Monad
    :param conditioner: A function that takes a function and returns a function that return a Monad.
    :type conditioner: Callable[[Callable[..., T]], Callable[..., "Monad[T]"]]
    :return: A decorator that wraps the given function in a Monad.
    :rtype: Callable[[Callable[..., T]], Callable[..., "Monad[T]"]]
    """

    def decorator(func: Callable[..., T]) -> Callable[..., "Monad[T]"]:
        """
        Decorator to wrap the result of a function in a Monad.

        :param func: The function to wrap in a Monad.
        :type func: Callable[..., T]
        :return: The wrapped function.
        :rtype: Callable[..., Monad[T]]
        """

        def wrapper(*arg, **kwargs) -> "Monad[T]":
            if conditioner:
                return conditioner(func)(*arg, **kwargs)
            return unit(m, func(*arg, **kwargs))

        try:
            wrapper.__annotations__ = func.__annotations__
        finally:
            wrapper.__name__ = func.__name__
            wrapper.__doc__ = func.__doc__
        return wrapper

    return decorator



from typing import Any, Callable, Generic, Optional, TypeVar
from fpylib.lazyness import lazy_eval

T = TypeVar("T")
S = TypeVar("S")


class Functor(Generic[T]):
    """
    Class Functor is a generic class that implements the Functor type class.


    :param value: The value of the functor.
    :type value: T
    :return: A functor with the given value.
    """

    def __init__(self, value: Optional[T] = None) -> None:
        """
        Initialize a functor with the given value.

        :param value: The value of the functor.
        :type value: Optional[T]
        :return: A functor with the given value.
        """

        def frozen_setattr(cls, __name: str, __value: Any) -> None:
            raise AttributeError("This object is not modifiable")

        object.__setattr__(self, "_Functor__value", value)
        object.__setattr__(self, "__setattr__", frozen_setattr)

    def get(self) -> T:
        """
        Get the value of the functor.

        :return: The value of the functor.
        """
        return self.__value

    def fmap(self, func: Callable[["Functor[T]"], "Functor[S]"]) -> "Functor[S]":
        """
        Apply a function to the value of the functor.

        :param func: The function to apply.
        :type func: Callable[["Functor[T]"], "Functor[S]"]
        :return: A functor with the value of the functor after applying the function.
        """
        return Functor(func(self.get()))


@lazy_eval
def fmap(func: Callable[[Functor[T]], Functor[S]], ft: Functor[T]) -> Functor[S]:
    """
    This function is a functor's fmap function.

    :param func: Callable[[Functor[T]], Functor[S]]
    :param ft: Functor[T]
    :return: Functor[S]
    """
    return ft.fmap(func)
